# -*- coding: utf-8 -*-

"""Topological methods for the system"""

import logging
from math import floor
import pprint  # noqa: F401

try:
    from openbabel import openbabel
except ModuleNotFoundError:
    print(
        "Please install openbabel using conda:\n"
        "     conda install -c conda-forge openbabel"
    )
    raise

logger = logging.getLogger(__name__)


class TopologyMixin:
    """A mixin for handling topology in a configuration."""

    def find_molecules(self, as_indices=False):
        """Find the separate molecules.

        Parameters
        ----------
        as_indices : bool = False
            Whether to return 0-based indices (True) or atom ids (False)

        Returns
        -------
        molecules : [[int]*n_molecules]
            A list of lists of atom ids or indices for the molecules
        """

        molecules = []

        atoms = self.atoms
        atom_ids = atoms.ids
        n_atoms = len(atom_ids)

        if n_atoms == 0:
            return molecules

        to_index = {j: i for i, j in enumerate(atom_ids)}
        neighbors = self.bonded_neighbors()
        visited = [False] * n_atoms
        while True:
            # Find first atom not yet visited
            try:
                index = visited.index(False)
            except ValueError:
                break
            visited[index] = True
            i = atom_ids[index]
            atoms = [i]
            next_atoms = neighbors[i]
            while len(next_atoms) > 0:
                tmp = []
                for i in next_atoms:
                    if not visited[to_index[i]]:
                        atoms.append(i)
                        visited[to_index[i]] = True
                        tmp.extend(neighbors[i])
                next_atoms = tmp
            molecules.append(sorted(atoms))
        if as_indices:
            return [[to_index[j] for j in js] for js in molecules]
        else:
            return molecules

    def bonded_neighbors(self, as_indices=False, first_index=0):
        """The atoms bonded to each atom in the system.

        Parameters
        ----------
        as_indices : bool = False
            Whether to return 0-based indices (True) or atom ids (False)
        first_index : int = 0
            The smallest index, e.g. 0 or 1

        Returns
        -------
        neighbors : {int: [int]} or [[int]] for indices
            list of atom ids for each atom id
        """
        neighbors = {}

        atoms = self.atoms
        bonds = self.bonds
        n_atoms = atoms.n_atoms

        if n_atoms == 0:
            if as_indices:
                return []
            else:
                return neighbors

        atom_ids = atoms.ids
        neighbors = {i: [] for i in atom_ids}

        if bonds.n_bonds > 0:
            for bond in bonds.bonds():
                i = bond["i"]
                j = bond["j"]
                neighbors[i].append(j)
                neighbors[j].append(i)

        if as_indices:
            # Convert to indices
            to_index = {j: i + first_index for i, j in enumerate(atom_ids)}
            result = [[]] * (n_atoms + first_index)
            for i, js in neighbors.items():
                result[to_index[i]] = sorted([to_index[j] for j in js])
            return result
        else:
            for i in neighbors:
                neighbors[i].sort()

            return neighbors

    def create_molecule_subsets(self):
        """Create a subset for each molecule in a configuration.

        Returns
        -------
        [int]
            The ids of the subsets, one per molecule.
        """
        # Find the molecules and the create the subsets if they don't exist.
        molecules = self.find_molecules()

        # Remove any previous subsets for this configuration
        subsets = self["subset"]
        tid = 1
        sids = subsets.find(tid)
        if len(sids) > 0:
            subsets.delete(sids)

        # Now create the new set.
        sids = []
        for atom_ids in molecules:
            sid = subsets.create(tid, atoms=atom_ids)
            sids.append(sid)

        return sids

    def create_molecule_templates(self, full_templates=True, create_subsets=True):
        """Create a template for each unique molecule in a configuration.

        By default also create subsets linking each template to the atoms
        of the molecules in the system.

        Parameters
        ----------
        full_templates : bool = True
            If true, create full templates by creating systems for the
            molecules.
        create_subsets : bool = True
            If true, create subsets linking the templates to the molecules.

        Returns
        -------
        [int] or [[int], [int]]
            The ids of the templates, or if create_subsets is True
            a two-element list containing the list of templates and
            list of subsets.
        """
        templates = self.system_db.templates

        # Find the molecules
        molecules = self.find_molecules()
        n_molecules = len(molecules)

        # And the molecule each atom is in
        atom_to_molecule = {}
        for molecule, atoms in enumerate(molecules):
            for atom in atoms:
                atom_to_molecule[atom] = molecule

        # The bonds in each molecule
        bonds_per_molecule = [[] for i in range(n_molecules)]
        for bond in self.bonds.bonds():
            i = bond["i"]
            j = bond["j"]
            order = bond["bondorder"]
            molecule = atom_to_molecule[i]
            bonds_per_molecule[molecule].append((i, j, order))

        # Get the canonical smiles for each molecule
        to_can = openbabel.OBConversion()
        to_can.SetOutFormat("can")
        ob_mol = openbabel.OBMol()
        ob_template = openbabel.OBMol()
        atnos = self.atoms.atomic_numbers
        xyzs = self.atoms.get_coordinates(fractionals=False)

        atom_index = {j: i for i, j in enumerate(self.atoms.ids)}

        new_subsets = {}
        sids = {}
        new_templates = []
        tids = []
        for molecule, atoms in enumerate(molecules):
            to_index = {j: i for i, j in enumerate(atoms)}
            molecule_atnos = [atnos[atom_index[i]] for i in atoms]

            ob_mol.Clear()
            for atom, atno in zip(atoms, molecule_atnos):
                ob_atom = ob_mol.NewAtom()
                ob_atom.SetAtomicNum(atno)
            bonds = []
            for i, j, order in bonds_per_molecule[molecule]:
                bonds.append((to_index[i], to_index[j], order))
                # 1-based indices in ob.
                ob_mol.AddBond(to_index[i] + 1, to_index[j] + 1, order)

            canonical = to_can.WriteString(ob_mol).strip()

            if full_templates:
                # See if a molecule template with the canonical smiles exists
                if templates.exists(canonical, "molecule"):
                    template = templates.get(canonical, category="molecule")
                else:
                    # Create a new system & configuration for the template
                    system_name = "template system " + canonical
                    if not self.system_db.system_exists(system_name):
                        system = self.system_db.create_system(
                            system_name, make_current=False
                        )
                        configuration = system.create_configuration(
                            canonical, make_current=False
                        )
                        cid = configuration.id

                        kwargs = {}
                        kwargs["atno"] = molecule_atnos
                        molecule_xyzs = [xyzs[atom_index[i]] for i in atoms]
                        kwargs["x"] = [x for x, y, z in molecule_xyzs]
                        kwargs["y"] = [y for x, y, z in molecule_xyzs]
                        kwargs["z"] = [z for x, y, z in molecule_xyzs]

                        ids = configuration.atoms.append(**kwargs)

                        kwargs = {}
                        kwargs["i"] = [ids[x] for x, _, _ in bonds]
                        kwargs["j"] = [ids[x] for _, x, _ in bonds]
                        kwargs["bondorder"] = [x for _, _, x in bonds]

                        configuration.bonds.append(**kwargs)

                        template = templates.create(
                            canonical, category="molecule", configuration=cid
                        )
            else:
                if templates.exists(canonical, "molecule"):
                    template = templates.get(canonical, category="molecule")
                else:
                    template = templates.create(canonical, category="molecule")

            if template.id not in tids:
                tids.append(template.id)
                new_templates.append(template)

            if create_subsets:
                if full_templates:
                    # Need to reorder the atoms to match the template atoms
                    # Prepare the OB molecule for the template
                    ob_template.Clear()
                    for atno in template.atoms.atomic_numbers:
                        ob_atom = ob_template.NewAtom()
                        ob_atom.SetAtomicNum(atno)

                    tatom_ids = template.atoms.ids
                    to_index = {j: i for i, j in enumerate(tatom_ids)}
                    for row in template.bonds.bonds():
                        i = to_index[row["i"]]
                        j = to_index[row["j"]]
                        order = row["bondorder"]
                        ob_template.AddBond(i + 1, j + 1, order)

                    # Get the mapping from template to molecule
                    query = openbabel.CompileMoleculeQuery(ob_template)
                    mapper = openbabel.OBIsomorphismMapper.GetInstance(query)
                    mapping = openbabel.vpairUIntUInt()
                    mapper.MapFirst(ob_mol, mapping)
                    reordered_atoms = [atoms[j] for i, j in mapping]

                    subset = self.subsets.create(template, reordered_atoms, tatom_ids)
                else:
                    subset = self.subsets.create(template, atoms)

                tid = template.id
                if tid not in sids:
                    sids[tid] = []
                    new_subsets[tid] = []
                sids[tid].append(subset.id)
                new_subsets[tid].append(subset)

        if create_subsets:
            return new_templates, new_subsets
        else:
            return new_templates

    def reimage_bonded_atoms(self, reimage_molecules=True):
        """Ensure that the atoms in a molecule are "near" each other.

        In a periodic system atoms can be translated by a unit cell without changing
        anything. However the bonds in a molecule need to account for such translations
        of atoms otherwise a bond may appear to be very long.

        It is often convenient physically move atoms by the correct cell translations
        to bring the atoms of a molecule close to each other. That is what this method
        does, moving all the atoms close to the first. Optionally the molecule is
        also moved to bring its geometric center into the primary unit cell, i.e. with
        fractional coordinates in the range [0..1).

        Parameters
        ----------
        reimage_molecules : bool = True
            Whether to move molecules into the primary unit cell.

        Returns
        -------
        bool
            True if the coordinates were changed.
        """
        if self.periodicity == 0:
            # Nothing to do.
            return False

        logger.log(0, f"Configuration {self.id} {self}")
        logger.log(0, self.atoms)
        logger.log(0, 10 * "+")
        # The bonds from each atom
        bonds = self.bonded_neighbors(as_indices=True)

        logger.debug("Bonds:")
        logger.debug(pprint.pformat(bonds))
        logger.debug("")

        # And coordinates as fractionals
        tmp = self.atoms.get_coordinates(fractionals=True)
        xyz = [[x, y, z] for x, y, z in tmp]

        visited = [False] * len(xyz)

        # Find the molecules
        molecules = self.find_molecules(as_indices=True)

        # Work through the atoms, see if bonded atoms need to be moved
        changed = False
        for atoms in molecules:
            logger.debug("")
            logger.debug(f"{atoms=}")
            # if len(atoms) > 1:
            #     atom = sorted(atoms)[0]
            #     logger.debug(f" --> {atom}")
            #     if self._image_helper(bonds, xyz, atom):
            #         changed = True

            changed = self._image_helper2(bonds, xyz, visited, atoms[0], changed)
        # Check temporarily.
        for iatom, jatoms in enumerate(bonds):
            xyzi = xyz[iatom]
            for jatom in jatoms:
                xyzj = xyz[jatom]
                for vi, vj in zip(xyzi, xyzj):
                    delta = floor(vj - vi + 0.5)
                    if delta != 0:
                        print(
                            f"Warning: {iatom} - {jatom} delta = {vj:.3f} - {vi:.3f} = "
                            f"{delta}"
                        )

        # Move the center of molecules into the primary cell, if requested.
        if reimage_molecules:
            for atoms in molecules:
                cx = cy = cz = 0.0
                for atom in atoms:
                    x, y, z = xyz[atom]
                    cx += x
                    cy += y
                    cz += z
                n = len(atoms)
                dx = int(cx / n)
                dy = int(cy / n)
                dz = int(cz / n)
                if dx != 0 or dy != 0 or dz != 0:
                    changed = True
                    for atom in atoms:
                        x, y, z = xyz[atom]
                        xyz[atom] = [x - dx, y - dy, z - dz]
        if changed:
            self.atoms.set_coordinates(xyz, fractionals=True)

        return changed

    def _image_helper(self, bonds, xyz, atom):
        """Translate all the bonded neighbors with larger index.

        Parameters
        ----------
        bonds : [[int]]
            The bonded neighbors for each atom.
        xyz : [[float * 3]]
            The fractional coordinates of the atoms.
        """
        logger.debug(f"       helper {atom}: {bonds[atom]}")
        changed = False
        xyzi = xyz[atom]
        for jatom in bonds[atom]:
            logger.debug(f"\t\t{atom=} {jatom}")
            if jatom < atom:
                continue
            xyzj = xyz[jatom]
            new_xyz = []
            shift = False
            for vi, vj in zip(xyzi, xyzj):
                delta = floor(vj - vi + 0.5)
                if delta == 0:
                    new_xyz.append(vj)
                else:
                    new_xyz.append(vj - delta)
                    shift = True
                logger.log(
                    0, f"{atom:3d} - {jatom:3d}: {xyzi} {xyzj} --> {new_xyz} {shift}"
                )
            if shift:
                changed = True
                logger.debug(
                    f"{atom:3d} - {jatom:3d}: {xyzi} {xyzj} --> {new_xyz} {shift}"
                )
                xyz[jatom] = new_xyz

        for jatom in bonds[atom]:
            if jatom < atom:
                continue
            if self._image_helper(bonds, xyz, jatom):
                changed = True

        return changed

    def reimage_molecules(self):
        """Reimage molecules into the primary unit cell.

        The molecules are moved to bring their geometric center into the primary unit
        cell, i.e. with fractional coordinates in the range [0..1).

        Returns
        -------
        bool
            True if the coordinates were changed.
        """
        if self.periodicity == 0:
            # Nothing to do.
            return False

        # And coordinates as fractionals
        xyz = self.atoms.get_coordinates(fractionals=True)

        # Find the molecules
        molecules = self.find_molecules(as_indices=True)

        # Move the centor of molecules into the primary cell, if requested.
        changed = False
        for atoms in molecules:
            cx = cy = cz = 0.0
            for atom in atoms:
                x, y, z = xyz[atom]
                cx += x
                cy += y
                cz += z
            n = len(atoms)
            dx = int(cx / n)
            dy = int(cy / n)
            dz = int(cz / n)
            if dx != 0 or dy != 0 or dz != 0:
                changed = True
                for atom in atoms:
                    x, y, z = xyz[atom]
                    xyz[atom] = [x - dx, y - dy, z - dz]
        if changed:
            self.atoms.set_coordinates(xyz, fractionals=True)

        return changed

    def _image_helper2(self, bonds, xyz, visited, iatom, changed):
        visited[iatom] = True
        xyzi = xyz[iatom]
        for jatom in bonds[iatom]:
            logger.debug(f"\t\t{iatom=} {jatom}")
            if visited[jatom]:
                continue
            new_xyz = []
            xyzj = xyz[jatom]
            shift = False
            for vi, vj in zip(xyzi, xyzj):
                delta = floor(vj - vi + 0.5)
                if delta == 0:
                    new_xyz.append(vj)
                else:
                    new_xyz.append(vj - delta)
                    changed = True
                    shift = True
            if shift:
                logger.debug(
                    f"{iatom:3d} - {jatom:3d}: {xyzi} {xyzj} --> {new_xyz} {shift}"
                )
            xyz[jatom] = new_xyz
            changed = self._image_helper2(bonds, xyz, visited, jatom, changed)
        return changed



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest  # noqa: F401

"""Tests for handling SMILES."""


def test_to_smiles(AceticAcid):
    """Create a SMILES string from a system"""
    correct = "CC(=O)O"
    smiles = AceticAcid.to_smiles()

    if smiles != correct:
        print(smiles)
    assert smiles == correct


def test_to_smiles_with_name(AceticAcid):
    """Create a SMILES string from a system"""
    correct = "CC(=O)O"
    smiles = AceticAcid.smiles

    if smiles != correct:
        print(smiles)
    assert smiles == correct


def test_to_canonical_smiles(AceticAcid):
    """Create a SMILES string from a system"""
    correct = "CC(=O)O"
    smiles = AceticAcid.to_smiles(canonical=True)

    if smiles != correct:
        print(smiles)
    assert smiles == correct


def test_several_molecules(CH3COOH_3H2O):
    """System with acetic acid and 3 waters"""
    system = CH3COOH_3H2O
    correct = "CC(=O)O.O.O.O"

    smiles = system.to_smiles()

    if smiles != correct:
        print(smiles)
    assert smiles == correct


def test_several_molecules_canonical(CH3COOH_3H2O):
    """System with acetic acid and 3 waters"""
    system = CH3COOH_3H2O
    correct = "CC(=O)O.O.O.O"

    smiles = system.to_smiles(canonical=True)

    if smiles != correct:
        print(smiles)
    assert smiles == correct


def test_from_smiles(configuration):
    """Create a configuration from SMILES"""
    correct = "CC(=O)O"
    configuration.from_smiles("OC(=O)C", name="acetic acid")
    result = configuration.to_smiles(canonical=True)

    if result != correct:
        print(result)

    assert result == correct



# -*- coding: utf-8 -*-

import logging

logger = logging.getLogger(__name__)


class _Symmetry(object):
    """A class to handle point and space group symmetry.

    :meta public:
    """

    def __init__(self, system_db, _id):
        """Initialize from the database.

        Parameters
        ----------
        system_db : SystemDB
            The SystemDB instance that we are working with.
        _id : int
            The id of this particular symmetry.
        """
        self._system_db = system_db
        self._id = _id

        super().__init__()



# -*- coding: utf-8 -*-

from collections import Counter
from functools import reduce
import logging
import math

from .atoms import _Atoms
from .bonds import _Bonds
from .cell import _Cell
from .cif import CIFMixin
from .configuration_properties import _ConfigurationProperties
from .molfile import MolFileMixin
from .openbabel import OpenBabelMixin
from .rdkit_ import RDKitMixin
from .pdb import PDBMixin
from .smiles import SMILESMixin
from .subsets import _Subsets
from .symmetry import _Symmetry
from .topology import TopologyMixin

logger = logging.getLogger(__name__)


class _Configuration(
    PDBMixin,
    MolFileMixin,
    CIFMixin,
    SMILESMixin,
    TopologyMixin,
    OpenBabelMixin,
    RDKitMixin,
    object,
):
    """A configuration (conformer) of a system.

    :meta public:
    """

    def __init__(self, _id, system_db):

        self._id = _id
        self._system_db = system_db

        self._name = None
        self._system = None
        self._atomset = None
        self._bondset = None
        self._cell_id = None
        self._charge = None
        self._coordinate_system = None
        self._spin_multiplicity = 0
        self._periodicity = None
        self._symmetry_id = None

    def __enter__(self):
        """Copy the tables to a backup for a 'with' statement."""
        self.system_db["configuration"].__enter__()
        self.atoms.__enter__()
        self.bonds.__enter__()
        # Avoid periodicity check
        _Cell(self).__enter__()
        return self

    def __exit__(self, etype, value, traceback):
        """Handle returning from a 'with' statement."""
        # Save the version, beacuse all of the subitems will update it.
        version = self.version

        self.system_db["configuration"].__exit__(etype, value, traceback)
        self.atoms.__exit__(etype, value, traceback)
        self.bonds.__exit__(etype, value, traceback)
        _Cell(self).__exit__(etype, value, traceback)

        if etype is None:
            self.version = version + 1

        return False

    @property
    def atoms(self):
        """The atoms for this configuration."""
        return _Atoms(self)

    @property
    def atomset(self):
        """The id of the atom set for this configuration."""
        if self._atomset is None:
            # Cache the atomset for this configuration
            self.cursor.execute(
                "SELECT atomset FROM configuration WHERE id = ?", (self.id,)
            )
            self._atomset = self.cursor.fetchone()[0]
        if self._atomset is None:
            # No atomset, so create one and update this configuration
            self._atomset = self.system_db["atomset"].append(n=1)[0]
            self.db.execute(
                "UPDATE configuration SET atomset = ? WHERE id = ?",
                (self._atomset, self.id),
            )
            self.db.commit()
        return self._atomset

    @atomset.setter
    def atomset(self, value):
        if self._atomset is None:
            # Cache the atomset for this configuration
            self.cursor.execute(
                "SELECT atomset FROM configuration WHERE id = ?", (self.id,)
            )
            tmp = self.cursor.fetchone()
            self._atomset = tmp[0]
        if self._atomset is None:
            # No atomset, so update this configuration
            self.db.execute(
                "UPDATE configuration SET atomset = ? WHERE id = ?",
                (value, self.id),
            )
            self.db.commit()
            self._atomset = value
        elif value == self._atomset:
            # Do nothing
            pass
        else:
            raise RuntimeError("The atomset is already set!")

    @property
    def bonds(self):
        """The bonds for this configuration."""
        return _Bonds(self)

    @property
    def bondset(self):
        """The id of the bond set for this configuration."""
        if self._bondset is None:
            # Cache the bondset for this configuration
            self.cursor.execute(
                "SELECT bondset FROM configuration WHERE id = ?", (self.id,)
            )
            self._bondset = self.cursor.fetchone()[0]
        if self._bondset is None:
            # No bondset, so create one and update this configuration
            self._bondset = self.system_db["bondset"].append(n=1)[0]
            self.db.execute(
                "UPDATE configuration SET bondset = ? WHERE id = ?",
                (self._bondset, self.id),
            )
            self.db.commit()
        return self._bondset

    @bondset.setter
    def bondset(self, value):
        if self._bondset is None:
            # Cache the bondset for this configuration
            self.cursor.execute(
                "SELECT bondset FROM configuration WHERE id = ?", (self.id,)
            )
            self._bondset = self.cursor.fetchone()[0]
        if self._bondset is None:
            # No bondset, so update this configuration
            self.db.execute(
                "UPDATE configuration SET bondset = ? WHERE id = ?",
                (value, self.id),
            )
            self.db.commit()
            self._bondset = value
        elif value == self._bondset:
            # No change
            pass
        else:
            raise RuntimeError("The bondset is already set!")

    @property
    def cell(self):
        """The periodic (unit) cell for this configuration.

        Raises
        ------
        TypeError
            If the system is not periodic.

        Returns
        -------
        _Cell
            The Cell object for this configuration.
        """
        if self.periodicity == 0:
            raise TypeError("The configuration is not periodic!")
        return _Cell(self)

    @property
    def cell_id(self):
        """The id of the cell in the cell table."""
        if self._cell_id is None:
            self.cursor.execute(
                "SELECT cell FROM configuration WHERE id = ?", (self.id,)
            )
            self._cell_id = self.cursor.fetchone()[0]
            if self._cell_id is None:
                self._cell_id = self.system_db["cell"].append(n=1)[0]
                sql = "UPDATE configuration SET cell = ? WHERE id = ?"
                self.db.execute(sql, (self._cell_id, self.id))
                self.db.commit()
        return self._cell_id

    @property
    def charge(self):
        """The charge of the system, 0, ±1, ±2, ±3, ..."""
        if self._charge is None:
            self.cursor.execute(
                "SELECT charge FROM configuration WHERE id = ?", (self.id,)
            )
            self._charge = self.cursor.fetchone()[0]
        return self._charge

    @charge.setter
    def charge(self, value):
        self.cursor.execute(
            "UPDATE configuration SET charge = ? WHERE id = ?", (value, self.id)
        )
        self.db.commit()
        self._charge = value

    @property
    def coordinate_system(self):
        """The coordinate system for this configuration."""
        if self._coordinate_system is None:
            self.cursor.execute(
                "SELECT coordinate_system FROM configuration WHERE id = ?", (self.id,)
            )
            self._coordinate_system = self.cursor.fetchone()[0]
        return self._coordinate_system

    @coordinate_system.setter
    def coordinate_system(self, value):
        if value.lower()[0] == "f":
            self.cursor.execute(
                "UPDATE configuration SET coordinate_system = 'fractional'"
                " WHERE id = ?",
                (self.id,),
            )
            self._coordinate_system = "fractional"
        else:
            self.cursor.execute(
                "UPDATE configuration SET coordinate_system = 'Cartesian'"
                " WHERE id = ?",
                (self.id,),
            )
            self._coordinate_system = "Cartesian"
        self.db.commit()

    @property
    def cursor(self):
        """The database connection."""
        return self.system_db.cursor

    @property
    def db(self):
        """The database connection."""
        return self.system_db.db

    @property
    def density(self):
        """Return the density of this configuration.

        Returns
        -------
        float
            The density of the cell.
        """
        if self.periodicity != 3:
            raise RuntimeError("Density is only defined for 3-D systems.")

        volume = self.volume
        mass = self.mass

        # converting from g/mol / Å^3 to g/cm^3
        return (mass / volume) * (1.0e24 / 6.02214076e23)

    @property
    def formula(self):
        """Return the chemical formula of the configuration.

        Returns a tuple with the formula, empirical formula and number of
        formula units (Z).

        Returns
        -------
        formulas : (str, str, int)
            The chemical formula, empirical formula and Z.
        """
        counts = Counter(self.atoms.symbols)

        # Order the elements ... Hill order if C, CH then alphabetical,
        # or if no C, then just alphabetically
        formula_list = []
        if "C" in counts:
            formula_list.append(("C", counts.pop("C")))
            if "H" in counts:
                formula_list.append(("H", counts.pop("H")))

        for element in sorted(counts.keys()):
            formula_list.append((element, counts[element]))

        counts = []
        for _, count in formula_list:
            counts.append(count)

        formula = []
        for element, count in formula_list:
            if count > 1:
                formula.append(f"{element}{count}")
            else:
                formula.append(element)

        # And the empirical formula
        Z = reduce(math.gcd, counts)
        empirical_formula_list = []
        for element, count in formula_list:
            empirical_formula_list.append((element, int(count / Z)))

        empirical_formula = []
        for element, count in empirical_formula_list:
            if count > 1:
                empirical_formula.append(f"{element}{count}")
            else:
                empirical_formula.append(element)

        return " ".join(formula), " ".join(empirical_formula), Z

    @property
    def id(self):
        """The id of this configuration."""
        return self._id

    @property
    def mass(self):
        """Return the total atomic mass.

        Returns
        -------
        float
            The summed atomic masses.
        """
        masses = self.atoms.atomic_masses

        return sum(masses)

    @property
    def spin_multiplicity(self):
        """The spin_multiplicity of the system, 0, 1, 2, 3, ..."""
        if self._spin_multiplicity == 0:
            self.cursor.execute(
                "SELECT spin_multiplicity FROM configuration WHERE id = ?", (self.id,)
            )
            multiplicity = self.cursor.fetchone()[0]
            if multiplicity is None or multiplicity == 0:
                n_electrons = sum(self.atoms.atomic_numbers) - self.charge
                if n_electrons % 2 == 0:
                    multiplicity = 1
                else:
                    multiplicity = 2
                self.spin_multiplicity = multiplicity
            else:
                self._spin_multiplicity = multiplicity
        return self._spin_multiplicity

    @spin_multiplicity.setter
    def spin_multiplicity(self, value):
        self.cursor.execute(
            "UPDATE configuration SET spin_multiplicity = ? WHERE id = ?",
            (value, self.id),
        )
        self.db.commit()
        self._spin_multiplicity = value

    @property
    def n_atoms(self) -> int:
        """The number of atoms.

        Returns
        -------
        int
            Number of atoms
        """
        return self.atoms.n_atoms

    @property
    def name(self):
        """The name of the configuration."""
        if self._name is None:
            self.cursor.execute(
                "SELECT name FROM configuration WHERE id = ?", (self.id,)
            )
            self._name = self.cursor.fetchone()[0]
        return self._name

    @name.setter
    def name(self, value):
        self.db.execute(
            "UPDATE configuration SET name = ? WHERE id = ?", (value, self.id)
        )
        self.db.commit()
        self._name = value

    @property
    def n_bonds(self) -> int:
        """The number of bonds.

        Returns
        -------
        int
            Number of bonds
        """
        return self.bonds.n_bonds

    @property
    def n_active_electrons(self):
        """The number of active electrons in the configuration"""
        self.cursor.execute(
            "SELECT n_active_electrons FROM configuration WHERE id = ?", (self.id,)
        )
        return self.cursor.fetchone()[0]

    @n_active_electrons.setter
    def n_active_electrons(self, value):
        self.cursor.execute(
            "UPDATE configuration SET n_active_electrons = ? WHERE id = ?",
            (value, self.id),
        )
        self.db.commit()

    @property
    def n_active_orbitals(self):
        """The number of active orbitals in the configuration"""
        self.cursor.execute(
            "SELECT n_active_orbitals FROM configuration WHERE id = ?", (self.id,)
        )
        return self.cursor.fetchone()[0]

    @n_active_orbitals.setter
    def n_active_orbitals(self, value):
        self.cursor.execute(
            "UPDATE configuration SET n_active_orbitals = ? WHERE id = ?",
            (value, self.id),
        )
        self.db.commit()

    @property
    def periodicity(self):
        """The periodicity of the system, 0, 1, 2 or 3"""
        if self._periodicity is None:
            self.cursor.execute(
                "SELECT periodicity FROM configuration WHERE id = ?", (self.id,)
            )
            self._periodicity = self.cursor.fetchone()[0]
        return self._periodicity

    @periodicity.setter
    def periodicity(self, value):
        if value < 0 or value > 3:
            raise ValueError("The periodicity must be between 0 and 3.")
        self.cursor.execute(
            "UPDATE configuration SET periodicity = ? WHERE id = ?", (value, self.id)
        )
        self.db.commit()
        self._periodicity = value

    @property
    def properties(self):
        """The class to handle the properties for this configuration."""
        return _ConfigurationProperties(self)

    @property
    def subsets(self):
        """The subsets"""
        return _Subsets(self)

    @property
    def state(self):
        """The electronic state of the configuration. Either a number or e.g. 2T1g
        for the second T1g state."""
        self.cursor.execute("SELECT state FROM configuration WHERE id = ?", (self.id,))
        return self.cursor.fetchone()[0]

    @state.setter
    def state(self, value):
        self.cursor.execute(
            "UPDATE configuration SET state = ? WHERE id = ?",
            (value, self.id),
        )
        self.db.commit()

    @property
    def symmetry(self):
        """The periodic (unit) symmetry for this configuration."""
        return _Symmetry(self.system_db, self.symmetry_id)

    @property
    def symmetry_id(self):
        """The id of the symmetry in the symmetry table."""
        if self._symmetry_id is None:
            self.cursor.execute(
                "SELECT symmetry FROM configuration WHERE id = ?", (self.id,)
            )
            self._symmetry_id = self.cursor.fetchone()[0]
        return self._symmetry_id

    @property
    def system(self):
        """The system that we belong to."""
        if self._system is None:
            self.cursor.execute(
                "SELECT system FROM configuration WHERE id = ?", (self.id,)
            )
            self._system = self.system_db.get_system(self.cursor.fetchone()[0])
        return self._system

    @property
    def system_db(self):
        """The system that we belong to."""
        return self._system_db

    @property
    def version(self):
        """The version of the system, incrementing from 0"""
        self.cursor.execute(
            "SELECT version FROM configuration WHERE id = ?", (self.id,)
        )
        return int(self.cursor.fetchone()[0])

    @version.setter
    def version(self, value):
        self.cursor.execute(
            "UPDATE configuration SET version = ? WHERE id = ?", (value, self.id)
        )
        self.db.commit()

    @property
    def volume(self):
        """Return the volume of this configuration.

        Returns
        -------
        float
            The volume of the cell.
        """
        if self.periodicity != 3:
            raise RuntimeError("Density is only defined for 3-D systems.")

        return self.cell.volume

    def clear(self) -> int:
        """Delete everything from the configuration."""
        # Delete the atoms
        self.atoms.delete("all")



# -*- coding: utf-8 -*-

"""
molsystem
A general implementation for molecular and periodic systems.
"""

# Bring up the classes so that they appear to be directly in
# the molsystem package.

import molsystem.elements  # noqa: F401
from molsystem.system_db import SystemDB  # noqa: F401
from molsystem.cell import Cell  # noqa: F401

# Handle versioneer
from ._version import get_versions

__author__ = """Paul Saxe"""
__email__ = "psaxe@molssi.org"
versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions



# -*- coding: utf-8 -*-

"""A dictionary-like object for holding atoms.
"""

from itertools import zip_longest
import logging
import pprint  # noqa: F401
from typing import Any, Dict, TypeVar

import numpy
import pandas

from molsystem import elements
from .column import _Column
from .table import _Table

Atoms_tp = TypeVar("Atoms_tp", "_Atoms", str, None)

logger = logging.getLogger(__name__)


def grouped(iterable, n):
    "s -> (s0,s1,s2,...sn-1), (sn,sn+1,sn+2,...s2n-1), (s2n,...s3n-1), ..."
    return zip_longest(*[iter(iterable)] * n)


class _Atoms(_Table):
    """:meta public:
    The atoms in a configuration of a system.

    Parameters
    ----------
    configuration : _Configuration
        The configuration of interest.
    """

    def __init__(self, configuration) -> None:

        self._configuration = configuration

        self._system_db = self._configuration.system_db
        self._system = None

        self._atomset = self._configuration.atomset

        self._atom_table = _Table(self.system_db, "atom")
        self._coordinates_table = _Table(self.system_db, "coordinates")

        super().__init__(self._system_db, "atom")

    def __enter__(self) -> Any:
        """Copy the tables to a backup for a 'with' statement."""
        self.system_db["atomset_atom"].__enter__()
        self.system_db["atom"].__enter__()
        return self

    def __exit__(self, etype, value, traceback) -> None:
        """Handle returning from a 'with' statement."""
        if etype is None:
            self.configuration.version = self.configuration.version + 1
        self.system_db["atomset_atom"].__exit__(etype, value, traceback)
        return self.system_db["atom"].__exit__(etype, value, traceback)

    def __delitem__(self, key) -> None:
        """Allow deletion of keys"""
        if key in self._atom_table.attributes:
            del self._atom_table[key]
        elif key in self._coordinates_table.attributes:
            del self._coordinates_table[key]

    def __iter__(self) -> iter:
        """Allow iteration over the object"""
        return iter([*self.attributes.keys()])

    def __len__(self) -> int:
        """The len() command"""
        return len(self.attributes)

    def __repr__(self) -> str:
        """The string representation of this object"""
        df = self.to_dataframe()
        return repr(df)

    def __str__(self) -> str:
        """The pretty string representation of this object"""
        df = self.to_dataframe()
        return df.to_string()

    def __contains__(self, item) -> bool:
        """Return a boolean indicating if a key exists."""
        # Normal the tablename is used as an identifier, so is quoted with ".
        # Here we need it as a string literal so strip any quotes from it.

        tmp_item = item.strip('"')
        return tmp_item in self.attributes

    def __eq__(self, other) -> Any:
        """Return a boolean if this object is equal to another"""
        diffs = self.diff(other)
        return len(diffs) == 0

    @property
    def ids(self):
        """The ids of the atoms."""
        return self.get_ids()

    @property
    def atomic_numbers(self):
        """The atomic numbers of the atoms.

        Returns
        -------
        [int]
            The atomic numbers.
        """
        return self.get_column_data("atno")

    @property
    def atomic_masses(self):
        """The atomic masses of the atoms.

        Returns
        -------
        [int]
            The atomic numbers.
        """

        if "mass" in self:
            result = self.get_column_data("mass")
        else:
            atnos = self.atomic_numbers
            result = elements.masses(atnos)
        return result

    @property
    def atomset(self):
        """The atomset for these atoms."""
        return self._atomset

    @property
    def attributes(self) -> Dict[str, Any]:
        """The definitions of the attributes.
        Combine the attributes of the atom and coordinates tables to
        make it look like a single larger table.
        """

        result = self._atom_table.attributes

        for key, value in self._coordinates_table.attributes.items():
            if key != "atom":  # atom key links the tables together, so ignore
                result[key] = value

        return result

    @property
    def configuration(self):
        """Return the configuration."""
        return self._configuration

    @property
    def coordinates(self):
        """The coordinates as list of lists."""
        return self.get_coordinates()

    @property
    def cursor(self):
        """The database connection."""
        return self.system_db.cursor

    @property
    def db(self):
        """The database connection."""
        return self.system_db.db

    @property
    def n_atoms(self) -> int:
        """The number of atoms in this configuration."""
        self.cursor.execute(
            "SELECT COUNT(*) FROM atomset_atom WHERE atomset = ?", (self.atomset,)
        )
        result = self.cursor.fetchone()[0]
        return result

    @property
    def symbols(self):
        """The element symbols for the atoms in this configuration.

        Returns
        -------
        [str]
            The element symbols
        """
        return elements.to_symbols(self.atomic_numbers)

    @property
    def system_db(self):
        """The system database that we belong to."""
        return self._system_db

    def add_attribute(
        self,
        name: str,
        coltype: str = "float",
        default: Any = None,
        notnull: bool = False,
        index: bool = False,
        pk: bool = False,
        references: str = None,
        on_delete: str = "cascade",
        on_update: str = "cascade",
        values: Any = None,
        configuration_dependent: bool = False,
    ) -> None:
        """Adds a new attribute.

        If the default value is None, you must always provide values wherever
        needed, for example when adding a row.

        Parameters
        ----------
        name : str
            The name of the attribute.
        coltype : str
            The type of the attribute (column). Must be one of 'int',
            'float', 'str' or 'byte'
        default : int, float, str or byte
            The default value for the attribute if no value is given.
        notnull : bool = False
            Whether the value must be non-null
        index : bool = False
            Whether to create an index on the column
        pk : bool = False
            Whether the column is the primry keys
        references : str = None
            If not null, the column is a foreign key for this table.
        on_delete : str = 'cascade'
            How to handle deletions of a foregin keys
        on_update : str = 'cascade'
            How to handle updates of a foregin key
        values : Any
            Either a single value or a list of values length 'nrows' of
            values to fill the column.
        configuration_dependent : bool = False
            Whether the attribute belongs with the coordinates (True)
            or atoms (False)

        Returns
        -------
            None
        """
        if configuration_dependent:
            self._coordinates_table.add_attribute(
                name,
                coltype=coltype,
                default=default,
                notnull=notnull,
                index=index,
                pk=pk,
                references=references,
                on_delete=on_delete,
                on_update=on_update,
                values=values,
            )
        else:
            self._atom_table.add_attribute(
                name,
                coltype=coltype,
                default=default,
                notnull=notnull,
                index=index,
                pk=pk,
                references=references,
                on_delete=on_delete,
                on_update=on_update,
                values=values,
            )

    def append(self, **kwargs: Dict[str, Any]) -> None:
        """Append one or more atoms.

        The keys give the field for the data. If an existing field is not
        mentioned, then the default value is used, unless the default is None,
        in which case an error is thrown. It is an error if there is not a
        field corrresponding to a key.
        """

        # Need to handle the elements specially. Can give atomic numbers,
        # or symbols. By construction the references to elements are identical
        # to their atomic numbers.

        if "symbol" in kwargs:
            symbols = kwargs.pop("symbol")
            kwargs["atno"] = elements.to_atnos(symbols)

        # How many new rows there are
        n_rows, lengths = self._get_n_rows(**kwargs)

        # Fill in the atom table
        data = {}
        for column in self._atom_table.attributes:
            if column != "id" and column in kwargs:
                data[column] = kwargs.pop(column)

        if len(data) == 0:
            data["atno"] = [None] * n_rows

        ids = self._atom_table.append(n=n_rows, **data)

        # Now append to the coordinates table
        configuration = self.configuration.id
        data = {"configuration": configuration, "atom": ids}
        for column in self._coordinates_table.attributes:
            if column != "id" and column in kwargs:
                data[column] = kwargs.pop(column)

        self._coordinates_table.append(n=n_rows, **data)

        # And to the atomset
        table = _Table(self.system_db, "atomset_atom")
        table.append(atomset=self.atomset, atom=ids)

        return ids

    def atoms(self, *args):
        """Return an iterator over the atoms.

        Parameters
        ----------
        args : [str]
            Added selection criteria for the SQL, one word at a time.

        Returns
        -------
        sqlite3.Cursor
            A cursor that returns sqlite3.Row objects for the atoms.
        """
        columns = self._columns()
        column_defs = ", ".join(columns)

        # What tables are requested in the extra arguments?
        tables = set()
        if len(args) > 0:
            atom_columns = [*self._atom_table.attributes]
            coordinates_columns = [*self._coordinates_table.attributes]
            for col, op, value in grouped(args, 3):
                if "." in col:
                    tables.add(col.split(".")[0])
                elif col in atom_columns:
                    tables.add("at")
                elif col in coordinates_columns:
                    tables.add("co")
                else:
                    raise ValueError(f"Column '{col}' is not available")

        # Build the query based on the tables needed
        sql = (
            f"SELECT {column_defs}"
            " FROM atomset_atom AS aa, atom AS at, coordinates AS co "
            "WHERE aa.atomset = ?"
            "  AND at.id = aa.atom"
            "  AND co.atom = at.id AND co.configuration = ?"
        )
        parameters = [self.atomset, self.configuration.id]

        # And any extra selection criteria
        if len(args) > 0:
            for col, op, value in grouped(args, 3):
                if op == "==":
                    op = "="
                sql += f' AND "{col}" {op} ?'
                parameters.append(value)

        logger.debug("atoms query:")
        logger.debug(sql)
        logger.debug(parameters)
        logger.debug("---")

        return self.db.execute(sql, parameters)

    def diff(self, other):
        """Difference between these atoms and another

        Parameters
        ----------
        other : _Atoms
            The other atoms to diff against

        Result
        ------
        result : Dict
            The differences, described in a dictionary
        """
        result = {}

        # Check the columns
        columns = self._columns()
        other_columns = other._columns()

        column_defs = ", ".join(columns)
        other_column_defs = ", ".join(other_columns)

        if columns == other_columns:
            column_def = column_defs
        else:
            added = columns - other_columns
            if len(added) > 0:
                result["columns added"] = list(added)
            deleted = other_columns - columns
            if len(deleted) > 0:
                result["columns deleted"] = list(deleted)

            in_common = other_columns & columns
            if len(in_common) > 0:
                column_def = ", ".join(in_common)
            else:
                # No columns shared
                return result

        # Need to check the contents of the tables. See if they are in the same
        # database or if we need to attach the other database temporarily.
        db = self.system_db
        other_db = other.system_db

        detach = False
        schema = self.schema
        if db.filename != other_db.filename:
            if db.is_attached(other_db):
                other_schema = db.attached_as(other_db)
            else:
                # Attach the other system_db in order to do comparisons.
                other_schema = self.system_db.attach(other_db)
                detach = True
        else:
            other_schema = other.schema

        atomset = self.atomset
        other_atomset = other.atomset

        changed = {}
        last = None
        sql = f"""
        SELECT * FROM
        (
          SELECT {column_def}
            FROM {other_schema}.atom AS at, {other_schema}.coordinates AS co
           WHERE co.atom = at.id
             AND at.id
              IN (
                  SELECT atom
                    FROM {other_schema}.atomset_atom
                   WHERE atomset = {other_atomset}
                  )
          EXCEPT
          SELECT {column_def}
            FROM {schema}.atom AS at, {schema}.coordinates AS co
           WHERE co.atom = at.id
             AND at.id
              IN (
                  SELECT atom
                    FROM {schema}.atomset_atom
                    WHERE atomset = {atomset}
                 )
        )
         UNION ALL
        SELECT * FROM
        (
          SELECT {column_def}
            FROM {schema}.atom AS at, {schema}.coordinates AS co
           WHERE co.atom = at.id
             AND at.id
              IN (
                  SELECT atom
                    FROM {schema}.atomset_atom
                   WHERE atomset = {atomset}
                 )
          EXCEPT
          SELECT {column_def}
            FROM {other_schema}.atom AS at, {other_schema}.coordinates AS co
           WHERE co.atom = at.id
             AND at.id
              IN (
                  SELECT atom
                    FROM {other_schema}.atomset_atom
                   WHERE atomset = {other_atomset}
                 )
        )
        ORDER BY id
        """

        for row in self.db.execute(sql):
            if last is None:
                last = row
            elif row["id"] == last["id"]:
                # changes = []
                changes = set()
                for k1, v1, v2 in zip(last.keys(), last, row):
                    if v1 != v2:
                        changes.add((k1, v1, v2))
                changed[row["id"]] = changes
                last = None
            else:
                last = row
        if len(changed) > 0:
            result["changed"] = changed

        # See about the rows added
        added = {}
        sql = f"""
        SELECT {column_defs}
          FROM {schema}.atom AS at, {schema}.coordinates AS co
         WHERE co.atom = at.id
           AND at.id
            IN (
                SELECT atom
                  FROM {schema}.atomset_atom
                 WHERE atomset = {atomset}
               )
           AND at.id
        NOT IN (
                SELECT atom
                  FROM {other_schema}.atomset_atom
                 WHERE atomset = {other_atomset}
               )
        """
        for row in self.db.execute(sql):
            added[row["id"]] = row[1:]

        if len(added) > 0:
            result["columns in added rows"] = row.keys()[1:]
            result["added"] = added

        # See about the rows deleted
        deleted = {}
        sql = f"""
        SELECT {other_column_defs}
          FROM {other_schema}.atom AS at, {other_schema}.coordinates AS co
         WHERE co.atom = at.id
           AND at.id
            IN (
                SELECT atom
                  FROM {other_schema}.atomset_atom
                 WHERE atomset = {other_atomset}
               )
           AND at.id
        NOT IN (
                SELECT atom
                  FROM {schema}.atomset_atom
                 WHERE atomset = {atomset}
               )
        """
        for row in self.db.execute(sql):
            deleted[row["id"]] = row[1:]

        if len(deleted) > 0:
            result["columns in deleted rows"] = row.keys()[1:]
            result["deleted"] = deleted

        # Detach the other database if needed
        if detach:
            self.system_db.detach(other_db)

        return result

    def get_as_dict(self, *args):
        """Return the atom data as a Python dictionary of lists.

        Parameters
        ----------
        args : [str]
            Added selection criteria for the SQL, one word at a time.

        Returns
        -------
        dict(str: [])
            A dictionary whose keys are the column names and values as lists
        """
        rows = self.atoms(*args)
        columns = [x[0] for x in rows.description]
        data = {key: [] for key in columns}
        for row in rows:
            for key, value in zip(columns, row):
                data[key].append(value)

        return data

    def get_ids(self, *args):
        """The ids of the selected atoms.

        Any extra arguments are triplets of column, operator, and value force
        additional selection criteria. The table names must be used in the column
        specification and are 'at' for the atom table and 'co' for the coordinate
        table.

        For example, if there are three arguments "at.atno", "=", "6" only the ids
        of the carbon atoms in the configuration will be returned.

        Parameters
        ----------
        args : [str]
            Further selection arguments, in sets of three:
            column, operator, and value, e.g. "at.atno = 6"

        Returns
        -------
        [int]
            The ids of the requested atoms.
        """

        # What tables are requested in the extra arguments?
        tables = set()
        if len(args) > 0:
            atom_columns = [*self._atom_table.attributes]
            coordinates_columns = [*self._coordinates_table.attributes]
            for col, op, value in grouped(args, 3):
                if "." in col:
                    tables.add(col.split(".")[0])
                elif col in atom_columns:
                    tables.add("at")
                elif col in coordinates_columns:
                    tables.add("co")
                else:
                    raise ValueError(f"Column '{col}' is not available")

        # Build the query based on the tables needed
        sql = "SELECT aa.atom FROM atomset_atom AS aa"
        if "at" in tables or "co" in tables:
            sql += ", atom AS at"
        if "co" in tables:
            sql += ", coordinates AS co"

        # The WHERE clauses bringing joining the tables
        sql += " WHERE aa.atomset = ?"
        parameters = [self.atomset]
        if "at" in tables or "co" in tables:
            sql += " AND at.id = aa.atom"
        if "co" in tables:
            sql += " AND co.atom = at.id AND co.configuration = ?"
            parameters.append(self.configuration.id)

        # And any extra selection criteria
        if len(args) > 0:
            for col, op, value in grouped(args, 3):
                if op == "==":
                    op = "="
                sql += f' AND "{col}" {op} ?'
                parameters.append(value)

        logger.debug("get_id query:")
        logger.debug(sql)
        logger.debug(parameters)
        logger.debug("---")

        return [x[0] for x in self.db.execute(sql, parameters)]

    def get_coordinates(self, fractionals=True, in_cell=False, as_array=False):
        """Return the coordinates optionally translated back into the principal
        unit cell.

        Parameters
        ----------
        fractionals : bool = True
            Return the coordinates as fractional coordinates for periodic
            systems. Non-periodic systems always use Cartesian coordinates.
        in_cell : bool, str = False
            Whether to translate the atoms into the unit cell, and if so
            whether to do so by molecule or just atoms.
        as_array : bool = False
            Whether to return the results as a numpy array or as a list of
            lists (the default).

        Returns
        -------
        abc : [N][float*3]
            The coordinates, either Cartesian or fractional
        """
        xyz = [[row["x"], row["y"], row["z"]] for row in self.atoms()]

        periodicity = self.configuration.periodicity
        if periodicity == 0:
            if as_array:
                return numpy.array(xyz)
            else:
                return xyz

        cell = self.configuration.cell

        if isinstance(in_cell, str) and "molecule" in in_cell:
            # Need fractionals...
            if self.configuration.coordinate_system == "Cartesian":
                UVW = cell.to_fractionals(xyz, as_array=True)
            elif not isinstance(xyz, numpy.ndarray):
                UVW = numpy.array(xyz)
            else:
                UVW = xyz

            molecules = self.configuration.find_molecules(as_indices=True)

            for indices in molecules:
                indices = numpy.array(indices)
                uvw_mol = numpy.take(UVW, indices, axis=0)
                center = numpy.average(uvw_mol, axis=0)
                delta = numpy.floor(center)
                uvw_mol -= delta
                numpy.put_along_axis(
                    UVW, numpy.expand_dims(indices, axis=1), uvw_mol, axis=0
                )
            if fractionals:
                if as_array:
                    return UVW
                else:
                    return UVW.tolist()
            else:
                return cell.to_cartesians(UVW, as_array=as_array)
        elif in_cell:
            # Need fractionals...
            if self.configuration.coordinate_system == "Cartesian":
                UVW = cell.to_fractionals(xyz, as_array=True)
            elif not isinstance(xyz, numpy.ndarray):
                UVW = numpy.array(xyz)
            else:
                UVW = xyz
            delta = numpy.floor(UVW)
            UVW -= delta
            if fractionals:
                if as_array:
                    return UVW
                else:
                    return UVW.tolist()
            else:
                return cell.to_cartesians(UVW, as_array=as_array)
        else:
            if fractionals:
                if self.configuration.coordinate_system == "Cartesian":
                    return cell.to_fractionals(xyz, as_array=as_array)
                elif as_array:
                    return numpy.array(xyz)
                else:
                    return xyz
            else:
                if self.configuration.coordinate_system == "fractional":
                    return cell.to_cartesians(xyz, as_array=as_array)
                elif as_array:
                    return numpy.array(xyz)
                else:
                    return xyz

    def get_n_atoms(self, *args):
        """Return the number of atoms meeting the cirteria.

        Parameters
        ----------
        args : [str]
            Added selection criteria for the SQL, one word at a time.

        Returns
        -------
        int
            The number of atoms matching the criteria.
        """
        # What tables are requested in the extra arguments?
        tables = set()
        if len(args) > 0:
            atom_columns = [*self._atom_table.attributes]
            coordinates_columns = [*self._coordinates_table.attributes]
            for col, op, value in grouped(args, 3):
                if "." in col:
                    tables.add(col.split(".")[0])
                elif col in atom_columns:
                    tables.add("at")
                elif col in coordinates_columns:
                    tables.add("co")
                else:
                    raise ValueError(f"Column '{col}' is not available")

        # Build the query based on the tables needed
        sql = "SELECT COUNT(*) FROM atomset_atom AS aa"
        if "at" in tables or "co" in tables:
            sql += ", atom AS at"
        if "co" in tables:
            sql += ", coordinates AS co"

        # The WHERE clauses bringing joining the tables
        sql += " WHERE aa.atomset = ?"
        parameters = [self.atomset]
        if "at" in tables or "co" in tables:
            sql += " AND at.id = aa.atom"
        if "co" in tables:
            sql += " AND co.atom = at.id AND co.configuration = ?"
            parameters.append(self.configuration.id)

        # And any extra selection criteria
        if len(args) > 0:
            for col, op, value in grouped(args, 3):
                if op == "==":
                    op = "="
                sql += f' AND "{col}" {op} ?'
                parameters.append(value)

        logger.debug("get_n_atoms query:")
        logger.debug(sql)
        logger.debug(parameters)
        logger.debug("---")

        self.cursor.execute(sql, parameters)
        return self.cursor.fetchone()[0]

    def set_coordinates(self, xyz, fractionals=True):
        """Set the coordinates to new values.

        Parameters
        ----------
        fractionals : bool = True
            The coordinates are fractional coordinates for periodic
            systems. Ignored for non-periodic systems.

        Returns
        -------
        None
        """
        as_array = isinstance(xyz, numpy.ndarray)

        x_column = self.get_column("x")
        y_column = self.get_column("y")
        z_column = self.get_column("z")

        xs = []
        ys = []
        zs = []

        periodicity = self.configuration.periodicity
        coordinate_system = self.configuration.coordinate_system
        if (
            periodicity == 0
            or (coordinate_system == "Cartesian" and not fractionals)
            or (coordinate_system == "fractional" and fractionals)
        ):
            if as_array:
                for x, y, z in xyz.tolist():
                    xs.append(x)
                    ys.append(y)
                    zs.append(z)
            else:
                for x, y, z in xyz:
                    xs.append(x)
                    ys.append(y)
                    zs.append(z)
        else:
            cell = self.configuration.cell
            if coordinate_system == "fractional":
                # Convert coordinates to fractionals
                for x, y, z in cell.to_fractionals(xyz):
                    xs.append(x)
                    ys.append(y)
                    zs.append(z)
            else:
                for x, y, z in cell.to_cartesians(xyz):
                    xs.append(x)
                    ys.append(y)
                    zs.append(z)
        x_column[0:] = xs
        y_column[0:] = ys
        z_column[0:] = zs

    def get_column(self, key: str) -> Any:
        """Get a Column object with the requested data

        Parameters
        ----------
        key : str
            The attribute to get.

        Returns
        -------
        Column
            A Column object containing the data.
        """
        if key in self._atom_table.attributes:
            sql = (
                f'SELECT at.rowid, at."{key}"'
                "   FROM atom as at, atomset_atom as aa"
                f" WHERE at.id = aa.atom AND aa.atomset = {self.atomset}"
            )
            return _Column(self._atom_table, key, sql=sql)
        elif key in self._coordinates_table.attributes:
            sql = (
                f'SELECT co.rowid, co."{key}"'
                "   FROM atom as at,"
                "        coordinates as co,"
                "        atomset_atom as aa"
                "  WHERE co.atom = at.id"
                f"   AND co.configuration = {self.configuration.id}"
                "    AND at.id = aa.atom"
                f"   AND aa.atomset = {self.atomset}"
            )
            return _Column(self._coordinates_table, key, sql=sql)
        else:
            raise KeyError(f"'{key}' not in atoms")

    def get_column_data(self, key: str) -> Any:
        """Return a column of data from the table.

        Parameters
        ----------
        key : str
            The attribute to get.

        Returns
        -------
        Column
            A Column object containing the data.
        """
        if key in self._atom_table.attributes:
            sql = (
                f'SELECT at."{key}"'
                "   FROM atom as at, atomset_atom as aa"
                f" WHERE at.id = aa.atom AND aa.atomset = {self.atomset}"
            )
            return [row[0] for row in self.db.execute(sql)]
        elif key in self._coordinates_table.attributes:
            sql = (
                f'SELECT co."{key}"'
                "   FROM atom as at,"
                "        coordinates as co,"
                "        atomset_atom as aa"
                "  WHERE co.atom = at.id"
                f"   AND co.configuration = {self.configuration.id}"
                "    AND at.id = aa.atom"
                f"   AND aa.atomset = {self.atomset}"
            )
            return [row[0] for row in self.db.execute(sql)]
        else:
            raise KeyError(f"'{key}' not in atoms")

    def _get_n_rows(self, **kwargs):
        """Get the total number of rows represented in the arguments."""
        n_rows = None

        lengths = {}
        for key, value in kwargs.items():
            if key not in self:
                raise KeyError(f'"{key}" is not an attribute of the atoms.')
            length = self.length_of_values(value)
            lengths[key] = length

            if n_rows is None:
                n_rows = 1 if length == 0 else length

            if length > 1 and length != n_rows:
                if n_rows == 1:
                    n_rows = length
                else:
                    raise IndexError(
                        'key "{}" has the wrong number of values, '.format(key)
                        + "{}. Should be 1 or the number of atoms ({}).".format(
                            length, n_rows
                        )
                    )
        return n_rows, lengths

    def length_of_values(self, values: Any) -> int:
        """Return the length of the values argument.

        Parameters
        ----------
        values : Any
            The values, which might be a string, single value, list, tuple,
            etc.

        Returns
        -------
        length : int
            The length of the values. 0 indicates a scalar
        """
        if isinstance(values, str):
            return 0
        else:
            try:
                return len(values)
            except TypeError:
                return 0

    def delete(self, atoms) -> int:
        """Delete the atoms listed

        Parameters
        ----------
        atoms : [int]
            The list of atoms to delete, or 'all' or '*'

        Returns
        -------
        None
        """
        # Delete the listed atoms, which will cascade to delete coordinates...
        if atoms == "all" or atoms == "*":
            sql = """
            DELETE FROM atom
             WHERE id IN (SELECT atom FROM atomset_atom WHERE atomset = ?)
            """
            parameters = (self.atomset,)
            self.db.execute(sql, parameters)
        else:
            sql = """
            DELETE FROM atom
             WHERE id = ?
               AND id IN (SELECT atom FROM atomset_atom WHERE atomset = ?)
            """
            parameters = [(i, self.atomset) for i in atoms]
            self.db.executemany(sql, parameters)

    def to_dataframe(self):
        """Return the contents of the table as a Pandas Dataframe."""
        data = {}
        rows = self.atoms()
        for row in rows:
            data[row[0]] = row[1:]

        columns = [x[0] for x in rows.description[1:]]

        df = pandas.DataFrame.from_dict(data, orient="index", columns=columns)

        return df

    def _columns(self):
        """The list of columns across the atom and coordinates table.

        Uses 'at' and 'co' as the shorthand for the fullt table names.
        """
        atom_columns = [*self._atom_table.attributes]
        coordinates_columns = [*self._coordinates_table.attributes]
        coordinates_columns.remove("atom")

        columns = [f'at."{x}"' for x in atom_columns]
        columns += [f'co."{x}"' for x in coordinates_columns]

        return columns


class _SubsetAtoms(_Atoms):
    """:meta public:
    The atoms in a subset.

    Parameters
    ----------
    configuration : _Configuration
        The configuration of interest.
    subset_id : int
        The id of the subset.

    Attributes
    ----------
    template_order : bool
        Whether to return atoms and properties in the order of the template
        if the template is full. Defaults to True.
    """

    def __init__(self, configuration, subset_id) -> None:

        self._sid = subset_id
        self.template_order = True

        # Caching
        self._template_id = None
        self._template = None

        super().__init__(configuration)

    def __eq__(self, other) -> Any:
        """Return a boolean if this object is equal to another"""
        raise NotImplementedError("Not implemented for subsets, yet!")

    @property
    def subset_id(self):
        """The subset for these atoms."""
        return self._sid

    @property
    def n_atoms(self) -> int:
        """The number of atoms in this configuration."""
        sql = """
        SELECT COUNT(*)
          FROM atomset_atom
         WHERE atomset = ?
           AND atom IN (SELECT atom FROM subset_atom WHERE subset = ?)
        """
        self.cursor.execute(sql, (self.atomset, self.subset_id))
        return self.cursor.fetchone()[0]

    @property
    def ids(self):
        """The ids of atoms in this subset."""
        sql = """
        SELECT atom
          FROM subset_atom
         WHERE subset = ?
        """
        if self.template.is_full and self.template_order:
            sql += "ORDER BY templateatom"

        return [x[0] for x in self.db.execute(sql, (self.subset_id,))]

    @property
    def template(self):
        """The template for this subset."""
        if self._template is None:
            self._template = self.system_db.templates.get(self.template_id)
        return self._template

    @property
    def template_id(self):
        """The id of the template for this subset."""
        if self._template_id is None:
            sql = "SELECT template FROM SUBSET WHERE id = ?"
            self.cursor.execute(sql, (self.subset_id,))
            self._template_id = self.cursor.fetchone()[0]
        return self._template_id

    def add(self, ids):
        """Add atoms to the subset.

        Parameters
        ----------
        ids : [int]
            The ids of the atoms to add. They are silently ignored if
            they are already in the subset.

        Raises
        ------
        ValueError
            If this subset has a full template or if an atom is not in the
            configuration.
        """
        # Check if the template has a template configuration
        if self.template.is_full:
            raise ValueError("Cannot add atoms to a subset for a full template.")

        # Remove any ids already in the subset
        atom_ids = set(self.ids)
        for aid in atom_ids & set(ids):
            ids.remove(aid)

        sa = self.system_db["subset_atom"]
        sa.append(subset=self.subset_id, atom=ids)

    def append(self, **kwargs: Dict[str, Any]) -> None:
        """Append one or more atoms.

        This is not allowed for a subset.

        Raises
        ------
        RuntimeError
            Adding atoms to a configuration is not allowed from a subset.
        """
        raise RuntimeError(
            "Adding atoms to a configuration is not allowed from a subset."
        )

    def atoms(self, *args):
        """Return an iterator over the atoms.

        Parameters
        ----------
        args : [str]
            Added selection criteria for the SQL, one word at a time.

        Returns
        -------
        sqlite3.Cursor
            A cursor that returns sqlite3.Row objects for the atoms.
        """
        columns = self._columns()
        column_defs = ", ".join(columns)

        sql = f"""
        SELECT {column_defs}
          FROM atom as at, coordinates as co, subset_atom as sa
         WHERE co.atom = at.id
           AND co.configuration = ?
           AND at.id = sa.atom
           AND sa.subset = ?
        """

        parameters = [self.configuration.id, self.subset_id]
        if len(args) > 0:
            for col, op, value in grouped(args, 3):
                if op == "==":
                    op = "="
                sql += f' AND "{col}" {op} ?'
                parameters.append(value)

        if self.template.is_full and self.template_order:
            sql += "ORDER BY sa.templateatom"

        return self.db.execute(sql, parameters)

    def get_n_atoms(self, *args):
        """Return the number of atoms meeting the cirteria.

        Parameters
        ----------
        args : [str]
            Added selection criteria for the SQL, one word at a time.

        Returns
        -------
        int
            The number of atoms matching the criteria.
        """
        sql = """
        SELECT COUNT(*)
          FROM atom as at, coordinates as co, subset_atom as sa
         WHERE co.atom = at.id
           AND co.configuration = ?
           AND at.id = sa.atom
           AND sa.subset = ?
        """
        parameters = [self.configuration.id, self.subset_id]
        if len(args) > 0:
            for col, op, value in grouped(args, 3):
                if op == "==":
                    op = "="
                sql += f' AND "{col}" {op} ?'
                parameters.append(value)

        self.cursor.execute(sql, parameters)
        return self.cursor.fetchone()[0]

    def diff(self, other):
        """Difference between these atoms and another

        Parameters
        ----------
        other : _Atoms
            The other atoms to diff against

        Result
        ------
        result : Dict
            The differences, described in a dictionary
        """
        raise NotImplementedError()

    def get_column(self, key: str) -> Any:
        """Get a Column object with the requested data

        Parameters
        ----------
        key : str
            The attribute to get.

        Returns
        -------
        Column
            A Column object containing the data.
        """
        if key in self._atom_table.attributes:
            sql = f"""
            SELECT at.rowid, at."{key}"
              FROM atom as at, subset_atom as sa
             WHERE at.id = sa.atom
               AND sa.subset = {self.subset_id}
            """
            if self.template.is_full and self.template_order:
                sql += "ORDER BY sa.templateatom"
            return _Column(self._atom_table, key, sql=sql)
        elif key in self._coordinates_table.attributes:
            sql = f"""
            SELECT co.rowid, co."{key}"
              FROM coordinates as co, subset_atom as sa
             WHERE co.atom = sa.atom
               AND co.configuration = {self.configuration.id}
               AND sa.subset = {self.subset_id}
            """
            if self.template.is_full and self.template_order:
                sql += "ORDER BY sa.templateatom"
            return _Column(self._coordinates_table, key, sql=sql)
        else:
            raise KeyError(f"'{key}' not in atoms")

    def get_column_data(self, key: str) -> Any:
        """Return a column of data from the table.

        Parameters
        ----------
        key : str
            The attribute to get.

        Returns
        -------
        Column
            A Column object containing the data.
        """
        if key in self._atom_table.attributes:
            sql = f"""
            SELECT at."{key}"
              FROM atom as at, subset_atom as sa
             WHERE at.id = sa.atom
               AND sa.subset = {self.subset_id}
            """
            if self.template.is_full and self.template_order:
                sql += "ORDER BY sa.templateatom"
            return [row[0] for row in self.db.execute(sql)]
        elif key in self._coordinates_table.attributes:
            sql = f"""
            SELECT co."{key}"
              FROM coordinates as co, subset_atom as sa
             WHERE co.atom = sa.atom
               AND co.configuration = {self.configuration.id}
               AND sa.subset = {self.subset_id}
            """
            if self.template.is_full and self.template_order:
                sql += "ORDER BY sa.templateatom"
            return [row[0] for row in self.db.execute(sql)]
        else:
            raise KeyError(f"'{key}' not in atoms")

    def get_ids(self, *args):
        """The ids of the atoms.

        Parameters
        ----------
        args : [str]
            Added selection criteria for the SQL, one word at a time.

        Returns
        -------
        [int]
            The ids of the requested atoms.
        """

        sql = """
        SELECT at.id
          FROM atom as at, coordinates as co, subset_atom as sa
         WHERE at.id = sa.atom
           AND co.configuration = ?
           AND sa.subset = ?
        """

        parameters = [self.configuration.id, self.subset_id]
        if len(args) > 0:
            for col, op, value in grouped(args, 3):
                if op == "==":
                    op = "="
                sql += f' AND "{col}" {op} ?'
                parameters.append(value)

        if self.template.is_full and self.template_order:
            sql += "ORDER BY sa.templateatom"

        return [x[0] for x in self.db.execute(sql, parameters)]

    def delete(self, atoms) -> int:
        """Delete the atoms listed

        Parameters
        ----------
        atoms : [int]
            The list of atoms to delete, or 'all' or '*'

        Returns
        -------
        None
        """
        raise NotImplementedError()

    def remove(self, ids):
        """Remove the given atoms from the subset.

        Parameters
        ----------
        ids : [int]
            The ids of the atoms to delete.

        Raises
        ------
        ValueError
            If this subset has a full template or if an atom is not in the
            subset.
        """
        # Check if the template has a template configuration
        if self.template.is_full:
            raise ValueError("Cannot add atoms to a subset for a full template.")

        # Check that the atoms are in this configuration!
        atom_ids = self.ids
        for aid in ids:
            if aid not in atom_ids:
                raise ValueError(f"Atom id={aid} is not in the subset.")

        parameters = [(_id, self.subset_id) for _id in ids]
        self.db.executemany(
            "DELETE FROM subset_atom WHERE atom=? AND subset=?", parameters
        )



# -*- coding: utf-8 -*-

import collections.abc
from itertools import zip_longest
import logging
import pandas

from .column import _Column

logger = logging.getLogger(__name__)

column_types = {"int": "INTEGER", "float": "REAL", "str": "TEXT", "bytes": "BLOB"}


def grouped(iterable, n):
    "s -> (s0,s1,s2,...sn-1), (sn,sn+1,sn+2,...s2n-1), (s2n,...s3n-1), ..."
    return zip_longest(*[iter(iterable)] * n)


class _Table(collections.abc.MutableMapping):
    """A dictionary-like object for holding tabular data

    This is a wrapper around an SQL table, with columns being the attributes.

    Attributes can be created, either from a list of predefined
    ones or by specifying the metadata required of an attribute. Attributes can
    also be deleted. See the method 'add_attribute' for more detail.

    Rows can be added ('append') or deleted ('delete').

    :meta public:
    """

    def __init__(self, system_db, table: str, other=None, logger=logger):
        self._system_db = system_db
        self.logger = logger
        if "." in table:
            self._schema, self._table = table.split(".")
        else:
            self._schema = "main"
            self._table = table
        if other is not None:
            self.copy(other)
        self._checkpoints = []

    def __enter__(self):
        """Copy the table to a backup for a 'with' statement."""
        n = len(self._checkpoints)
        checkpoint = f"temp.{self._table}_checkpoint_{n}"
        self.db.execute(f"CREATE TEMP TABLE {checkpoint} AS SELECT * FROM {self.table}")
        self.db.commit()
        self._checkpoints.append(checkpoint)
        return self

    def __exit__(self, etype, value, traceback) -> None:
        """Handle returning from a 'with' statement."""
        checkpoint = self._checkpoints.pop()

        if etype is None:
            self.db.commit()

            # Log the changes
            previous = _Table(self.system_db, checkpoint)
            diffs = self.diff(previous)
            if diffs is not None:
                self.logger.debug(diffs)
        else:
            # Delete everything in the table and copy the checkpoint back.
            self.db.execute("PRAGMA foreign_keys = OFF")
            self.clear()
            self.db.execute(f"INSERT INTO {self.table} SELECT * FROM {checkpoint}")
            self.db.execute("PRAGMA foreign_keys = ON")
            self.db.commit()

        # Drop the temporary table
        self.db.execute(f"DROP TABLE {checkpoint}")
        self.db.commit()

        return False

    def __getitem__(self, key):
        """Allow [] to access the data!"""
        return self.get_column(key)

    def __setitem__(self, key, value) -> None:
        """Allow x[key] access to the data"""
        self[key][0:] = value

    def __delitem__(self, key) -> None:
        """Allow deletion of keys"""
        # The easy way, which is not supported in SQLite :-(
        # self.cursor.execute(f'ALTER TABLE {self.table} DROP {key}')

        columns = set(self.attributes)
        columns.remove(key)
        column_def = ", ".join(columns)

        # Need the unquoted version!
        table = self._table

        sql = f"""
        -- disable foreign key constraint check
        PRAGMA foreign_keys=off;

        -- Here you can drop column
        CREATE TABLE {self.schema}."tmp_{table}"
           AS SELECT {column_def} FROM {self.table};

        -- drop the table
        DROP TABLE {self.table};

        -- rename the new_table to the table
        ALTER TABLE {self.schema}."tmp_{table}"
          RENAME TO "{self._table}";

        -- enable foreign key constraint check
        PRAGMA foreign_keys=on;
        """
        self.db.executescript(sql)
        self.db.commit()

    def __iter__(self) -> iter:
        """Allow iteration over the object"""
        return iter([*self.attributes.keys()])

    def __len__(self) -> int:
        """The len() command"""
        return len(self.attributes)

    def __repr__(self) -> str:
        """The string representation of this object"""
        df = self.to_dataframe()
        return repr(df)

    def __str__(self) -> str:
        """The pretty string representation of this object"""
        df = self.to_dataframe()
        return str(df)

    def __contains__(self, item) -> bool:
        """Return a boolean indicating if a key exists."""
        # Normal the tablename is used as an identifier, so is quoted with ".
        # Here we need it as a string literal so strip any quotes from it.

        tmp_item = item.strip('"')
        return tmp_item in self.attributes

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        # Check numbers of rows
        if self.n_rows != other.n_rows:
            return False

        # Check the columns
        attributes = self.attributes
        columns = set(attributes)

        other_attributes = other.attributes
        other_columns = set(other_attributes)

        if columns != other_columns:
            return False

        is_same = True
        # Need to check the contents of the tables. See if they are in the same
        # database or if we need to attach the other database temporarily.
        db = self.system_db
        other_db = other.system_db

        detach = False
        if db.filename != other_db.filename:
            if db.is_attached(other_db):
                schema = db.attached_as(other_db)
            else:
                # Attach the other system_db in order to do comparisons.
                schema = self.system_db.attach(other_db)
                detach = True
            other_table = f'"{schema}"."{other._table}"'
        else:
            other_table = other.table
        table = self.table

        self.cursor.execute(
            f"""
            SELECT COUNT(*) FROM
            (SELECT rowid, * FROM {other_table}
            EXCEPT
            SELECT rowid, * FROM {table})
            """
        )
        if self.cursor.fetchone()[0] != 0:
            is_same = False
        else:
            self.cursor.execute(
                f"""
                SELECT COUNT(*) FROM
                (SELECT rowid, * FROM {table}
                EXCEPT
                SELECT rowid, * FROM {other_table})
                """
            )
            if self.cursor.fetchone()[0] != 0:
                is_same = False

        # Detach the other database if needed
        if detach:
            self.system_db.detach(other_db)

        return is_same

    @property
    def system_db(self):
        """The system_db that we belong to."""
        return self._system_db

    @property
    def db(self):
        """The database connection."""
        return self.system_db.db

    @property
    def cursor(self):
        """The database connection."""
        return self.system_db.cursor

    @property
    def schema(self):
        """The database schema the table is in."""
        return f'"{self._schema}"'

    @property
    def table(self) -> str:
        """The name of this table."""
        return f'"{self._schema}"."{self._table}"'

    @property
    def n_rows(self) -> int:
        """The number of rows in the table."""
        self.cursor.execute(f"SELECT COUNT(*) FROM {self.table}")
        result = self.cursor.fetchone()[0]
        return result

    @property
    def attributes(self):
        """The definitions of the attributes."""
        result = {}
        for row in self.db.execute(
            "   SELECT *"
            f"    FROM {self.schema}.pragma_table_info('{self._table}')"
            " ORDER BY 'cid'"
        ):
            result[row["name"]] = {
                "type": row["type"],
                "notnull": bool(row["notnull"]),
                "default": row["dflt_value"],
                "primary key": bool(row["pk"]),
            }
        for row in self.db.execute(f"PRAGMA foreign_key_list('{self.table}')"):
            if row["to"] is None:
                result[row["from"]]["fk"] = row["table"]
            else:
                result[row["from"]]["fk"] = f"{row['table']}.{row['to']}"

        return result

    # @property
    # def version(self):
    #     return self.system_db.version

    def add_attribute(
        self,
        name: str,
        coltype: str = "float",
        default=None,
        notnull: bool = False,
        index: bool = False,
        pk: bool = False,
        references: str = None,
        on_delete: str = "cascade",
        on_update: str = "cascade",
        values=None,
    ) -> None:
        """Adds a new attribute.

        If the default value is None, you must always provide values wherever
        needed, for example when adding a row.

        Parameters
        ----------
        name : str
            The name of the attribute.
        coltype : str
            The type of the attribute (column). Must be one of 'int',
            'float', 'str' or 'byte'
        default : int, float, str or byte
            The default value for the attribute if no value is given.
        notnull : bool = False
            Whether the value must be non-null
        index : bool = False
            Whether to create an index on the column
        pk : bool = False
            Whether the column is the primry keys
        references : str = None
            If not null, the column is a foreign key for this table.
        on_delete : str = 'cascade'
            How to handle deletions of a foregin keys
        on_update : str = 'cascade'
            How to handle updates of a foregin key
        values : Any
            Either a single value or a list of values length 'nrows' of
            values to fill the column.

        Returns:
            None
        """
        # Does the table exist?
        if self.table in self.system_db:
            table_exists = True
            if pk:
                raise ValueError(
                    "The primary key can only be set on the first attribute "
                    "of a table that does not yet exist."
                )
        else:
            table_exists = False
            if pk:
                notnull = True

        if name in self:
            raise RuntimeError(
                "{} attribute '{}' is already defined!".format(
                    self.__class__.__name__, name
                )
            )

        # not null columns must have defaults
        if not pk and notnull and default is None:
            raise ValueError(f"Not null attributes must have defaults: {name}")

        # see if the values are given
        if values is not None:
            length = self.length_of_values(values)

            if length > 1 and length != self.n_rows:
                raise IndexError(
                    f"The number of values given, {length}, must be either 1, "
                    f"or the number of rows in {self.table}: {self.n_rows}"
                )

        # Create the column
        parameters = []
        if coltype in column_types:
            column_type = column_types[coltype]
        else:
            column_type = coltype
        column_def = f'"{name}" {column_type}'
        if default is not None:
            if coltype == "str":
                column_def += f" DEFAULT '{default}'"
            else:
                column_def += f" DEFAULT {default}"
        if notnull:
            column_def += " NOT NULL"
        if references is not None:
            column_def += f" REFERENCES {references}"
            if on_delete is not None and on_delete != "":
                column_def += f" ON DELETE {on_delete}"
            if on_update is not None and on_update != "":
                column_def += f" ON UPDATE {on_update}"

        if table_exists:
            self.cursor.execute(
                f"ALTER TABLE {self.table} ADD {column_def}", parameters
            )
        else:
            if pk:
                column_def += " PRIMARY KEY"
            self.cursor.execute(f"CREATE TABLE {self.table} ({column_def})", parameters)

        if index == "unique":
            self.cursor.execute(
                f'CREATE UNIQUE INDEX {self.schema}."idx_{name}"'
                f'  ON "{self._table}"'
                f"  ('{name}')"
            )
        elif index:
            self.cursor.execute(
                f'CREATE INDEX {self.schema}."idx_{name}"'
                f'  ON "{self._table}"'
                f"  ('{name}')"
            )

        if values is not None:
            self[name] = values

        self.db.commit()

    def append(self, n=None, **kwargs):
        """Append one or more rows

        The keywords are the names of attributes and the value to use.
        The default value for any attributes not given is used unless it is
        'None' in which case an error is thrown. It is an error if there is not
        an exisiting attribute corresponding to any given as arguments.

        Parameters
        ----------
        kwargs : keyword arguments
            Giving existing attributes and values.

        Returns
        -------
        [int]
            The ids of the created rows.
        """

        if len(kwargs) > 0:
            n_rows, lengths = self._get_n_rows(**kwargs)
        else:
            lengths = {}
            if n is not None:
                n_rows = n
            else:
                n_rows = 1

        if n is not None:
            if n_rows != 1 and n_rows != n:
                raise RuntimeError(
                    f"Requested number of rows ({n}) not compatible with the "
                    f"length of the data ({n_rows})."
                )
            n_rows = n

        # Check that any missing attributes have defaults
        attributes = self.attributes
        for key in attributes:
            if not attributes[key]["notnull"] or attributes[key]["primary key"]:
                continue

            if key not in kwargs:
                if (
                    "default" not in attributes[key]
                    or attributes[key]["default"] is None
                ):
                    raise KeyError(
                        "There is no default for attribute "
                        "'{}'. You must supply a value".format(key)
                    )

        # Add id's if needed
        if "id" not in kwargs and "id" in attributes:
            self.cursor.execute(f"SELECT MAX(id) FROM {self.table}")
            last_id = self.cursor.fetchone()[0]
            if last_id is None:  # Table is empty
                last_id = 0
            kwargs["id"] = [*range(last_id + 1, last_id + n_rows + 1)]
            lengths["id"] = n_rows

        # All okay, so proceed.
        values = {}
        for key, value in kwargs.items():
            if lengths[key] == 0:
                values[key] = [value] * n_rows
            elif lengths[key] == 1:
                values[key] = [value[0]] * n_rows
            else:
                values[key] = value

        parameters = []
        for row in range(n_rows):
            line = []
            for key, value in values.items():
                line.append(value[row])
            parameters.append(line)

        columns = '"' + '", "'.join(kwargs.keys()) + '"'
        places = ", ".join(["?"] * len(values.keys()))

        self.cursor.executemany(
            f"INSERT INTO {self.table} ({columns}) VALUES ({places})", parameters
        )

        self.db.commit()

        if "id" in kwargs:
            return kwargs["id"]

    def clear(self):
        """Delete all rows from the table."""
        self.db.execute(f"DELETE FROM {self.table}")

    def diff(self, other):
        """Difference between this table and another

        Parameters
        ----------
        other : _Table
            The other table to diff against

        Result
        ------
        result : dict
            The differences, described in a dictionary
        """
        self.logger.debug("diffs for {self.table}")
        result = {}
        summary = {}

        # Check the columns
        attributes = self.attributes
        columns = set(attributes)

        other_attributes = other.attributes
        other_columns = set(other_attributes)

        if columns == other_columns:
            column_def = "rowid, *"
        else:
            added = columns - other_columns
            if len(added) > 0:
                self.logger.debug(f"      columns added: {', '.join(sorted(added))}")
                result["columns added"] = added
                summary["columns add"] = added
            deleted = other_columns - columns
            if len(deleted) > 0:
                self.logger.debug(f"     columns deleted: {', '.join(sorted(deleted))}")
                result["columns deleted"] = deleted
                summary["columns deleted"] = deleted

            in_common = other_columns & columns
            if len(in_common) > 0:
                column_def = "rowid, " + ", ".join([f'"{x}"' for x in in_common])
            else:
                # No columns shared
                return result

        # Need to check the contents of the tables. See if they are in the same
        # database or if we need to attach the other database temporarily.
        db = self.system_db
        other_db = other.system_db

        detach = False
        if db.filename != other_db.filename:
            if db.is_attached(other_db):
                schema = db.attached_as(other_db)
            else:
                # Attach the other system_db in order to do comparisons.
                schema = self.system_db.attach(other_db)
                detach = True
            other_table = f'"{schema}"."{other._table}"'
        else:
            other_table = other.table
        table = self.table

        changed = {}
        last = None
        sql = f"""
        SELECT {column_def} FROM
         (
          SELECT {column_def} FROM {other_table}
           EXCEPT
          SELECT {column_def} FROM {table}
         )
        UNION ALL
        SELECT {column_def} FROM
         (
          SELECT {column_def} FROM {table}
           EXCEPT
          SELECT {column_def} FROM {other_table}
         )
        ORDER BY rowid
        """
        changed_columns = set()
        for row in self.db.execute(sql):
            if last is None:
                last = row
            elif row["rowid"] == last["rowid"]:
                changes = set()
                for k1, v1, v2 in zip(last.keys(), last, row):
                    if v1 != v2:
                        changed_columns.add(k1)
                        changes.add((k1, v1, v2))
                changed[row["rowid"]] = changes
                last = None
            else:
                last = row
        if len(changed) > 0:
            self.logger.debug(
                f"    columns changed: {', '.join(sorted(changed_columns))}"
            )
            result["changed"] = changed
            summary["columns changed"] = changed_columns

        # See about the rows added
        added = {}
        if "id" in self:
            for row in self.db.execute(
                f"""
                SELECT * FROM {table}
                WHERE rowid NOT IN (SELECT rowid FROM {other_table})
                """
            ):
                added[row["id"]] = row[1:]
        else:
            for row in self.db.execute(
                f"""
                SELECT rowid, * FROM {table}
                WHERE rowid NOT IN (SELECT rowid FROM {other_table})
                """
            ):
                added[row["rowid"]] = row[1:]

        if len(added) > 0:
            self.logger.debug(f"         rows added: {len(added)}")
            result["columns in added rows"] = row.keys()[1:]
            result["added"] = added
            summary["rows added"] = len(added)

        # See about the rows deleted
        deleted = {}
        if "id" in self:
            for row in self.db.execute(
                f"""
                SELECT * FROM {other_table}
                WHERE rowid NOT IN (SELECT rowid FROM {table})
                """
            ):
                deleted[row["id"]] = row[1:]
        else:
            for row in self.db.execute(
                f"""
                SELECT rowid, * FROM {other_table}
                WHERE rowid NOT IN (SELECT rowid FROM {table})
                """
            ):
                deleted[row["rowid"]] = row[1:]

        if len(deleted) > 0:
            self.logger.debug(f"       rows deleted: {len(deleted)}")
            result["columns in deleted rows"] = row.keys()[1:]
            result["deleted"] = deleted
            summary["rows deleted"] = len(deleted)

        # Detach the other database if needed
        if detach:
            self.system_db.detach(other_db)

        if len(result) == 0:
            self.logger.debug("    no changes")
            result = None
        else:
            result["summary"] = summary

        return result

    def get_as_dict(self, *args):
        """Return the table data as a Python dictionary of lists.

        Parameters
        ----------
        args : [str]
            Added selection criteria for the SQL, one word at a time.

        Returns
        -------
        dict(str: [])
            A dictionary whose keys are the column names and values as lists
        """
        rows = self.rows(*args)
        columns = [x[0] for x in rows.description]
        data = {key: [] for key in columns}
        for row in rows:
            for key, value in zip(columns, row):
                data[key].append(value)

        return data

    def get_column(self, column, sql=None, where=""):
        """Return a column of data from the table.

        Parameters
        ----------
        column : str
            The name of the column.
        sql : str = None
            Optional SQL statement to get the rows.
        where : str = ''
            Optional WHERE clause when getting the rows if no SQL is given.
            Should be the complete clause, starting with 'WHERE ...'
        """
        return _Column(self, column=column, sql=sql, where=where)

    def get_column_data(self, column, sql=None, where=""):
        """Return a column of data from the table as a list.

        Parameters
        ----------
        column : str
            The name of the column.
        sql : str = None
            Optional SQL statement to get the rows.
        where : str = ''
            Optional WHERE clause when getting the rows if no SQL is given.
            Should be the complete clause, starting with 'WHERE ...'
        """
        result = []
        if sql is None:
            sql = f"SELECT {column} FROM {self.table} {where}"

        for row in self.db.execute(sql):
            result.append(row[0])
        return result

    def delete(self, *args):
        """Delete rows matching the selection."""
        if len(args) == 0:
            return self.db.execute(f"DELETE FROM {self.table}")

        sql = f"DELETE FROM {self.table} WHERE"

        parameters = []
        clause = []
        for col, op, value in grouped(args, 3):
            if op == "==":
                op = "="
            clause.append(f'"{col}" {op} ?')
            parameters.append(value)

        sql += " AND ".join(clause)
        return self.db.execute(sql, parameters)

    def rows(self, *args):
        """Return an iterator over the rows."""
        if len(args) == 0:
            return self.db.execute(f"SELECT * FROM {self.table}")

        sql = f"SELECT * FROM {self.table} WHERE"

        parameters = []
        clause = []
        for col, op, value in grouped(args, 3):
            if op == "==":
                op = "="
            clause.append(f'"{col}" {op} ?')
            parameters.append(value)

        sql += " AND ".join(clause)
        return self.db.execute(sql, parameters)

    def _get_n_rows(self, **kwargs):
        """Get the total number of rows represented in the arguments."""
        n_rows = None

        lengths = {}
        for key, value in kwargs.items():
            if key not in self:
                raise KeyError(
                    f'"{key}" is not an attribute of the table ' f"{self.table}!"
                )
            length = self.length_of_values(value)
            lengths[key] = length

            if n_rows is None:
                n_rows = 1 if length == 0 else length

            if length > 1 and length != n_rows:
                if n_rows == 1:
                    n_rows = length
                else:
                    raise IndexError(
                        'key "{}" has the wrong number of values, '.format(key)
                        + "{}. Should be 1 or the number of rows in {} ({}).".format(
                            length, self.table, n_rows
                        )
                    )
        return n_rows, lengths

    def copy(self, other):
        """Replace this object with a copy of another."""
        if self.table in self.system_db:
            del self.system_db[self.table]

        # Need the contents of the tables. See if they are in the same
        # database or if we need to attach the other database temporarily.

        name = self.system_db.nickname
        other_name = other.system_db.nickname
        detach = False
        if name != other_name:
            if not self.system_db.is_attached(other_name):
                # Attach the other system_db in order to do comparisons.
                self.system_db.attach(other.system_db)
                detach = True
            other_table = f'"{other_name}"."{other._table}"'
        else:
            other_table = other.table
        table = self.table

        self.cursor.execute(f"CREATE TABLE {table} AS SELECT * FROM {other_table}")
        self.db.commit()

        # Detach the other database if needed
        if detach:
            self.system_db.detach(other.system_db)

    def length_of_values(self, values):
        """Return the length of the values argument.

        Parameters
        ----------
        values : Any
            The values, which might be a string, single value, list, tuple,
            etc.

        Returns
        -------
        length : int
            The length of the values. 0 indicates a scalar
        """
        if isinstance(values, str):
            return 0
        else:
            try:
                return len(values)
            except TypeError:
                return 0

    def to_dataframe(self):
        """Return the contents of the table as a Pandas Dataframe."""
        data = {}
        for line in self.cursor.execute(f"SELECT rowid, * FROM {self.table}"):
            data[line[0]] = line[1:]
        columns = [x[0] for x in self.cursor.description[1:]]

        df = pandas.DataFrame.from_dict(data, orient="index", columns=columns)

        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests of subsets."""

import pprint  # noqa: F401

import pytest  # noqa: F401


def test_construction(CH3COOH_3H2O):
    """Simplest test that we can make a new subsets object"""
    configuration = CH3COOH_3H2O
    subsets = configuration.subsets
    assert str(type(subsets)) == "<class 'molsystem.subsets._Subsets'>"


def test_n_subsets(CH3COOH_3H2O):
    """Test getting the number of subsets."""
    configuration = CH3COOH_3H2O
    subsets = configuration.subsets
    assert subsets.n_subsets == 0

    db = configuration.system_db
    templates = db.templates
    tid = templates.create(name="H2O", category="molecule")

    subset = subsets.create(template=tid)
    assert subsets.n_subsets == 1
    assert subset.id == 1

    ids = subsets.get_ids(tid)
    assert ids == [1]


def test_simple_subsets(simple_templates):
    """Test adding atoms to a simple subset."""
    answer = [6, 1, 1, 1, 6, 8, 8, 1]

    db = simple_templates
    templates = db.templates
    acy = templates.get("acy", "molecule")
    hoh = templates.get("hoh", "molecule")

    system = db.get_system("acetic acid")
    configuration = system.configuration
    subsets = configuration.subsets

    subsets.create(acy, atoms=[*range(1, 9)])
    subset = subsets.get(acy)[0]
    assert subset.atoms.n_atoms == 8

    assert subset.atoms.atomic_numbers == answer

    subsets.create(hoh, atoms=[9, 10, 11])
    subsets.create(hoh, atoms=[12, 13, 14])
    # for this one add the atoms later
    last = subsets.create(hoh)

    waters = subsets.get(hoh)

    # adding atoms to last water
    last.atoms.add([15, 16, 17])

    for water in waters:
        assert water.atoms.atomic_numbers == [8, 1, 1]
